Package: highr
Type: Package
Title: Syntax Highlighting for R Source Code
Version: 0.9
Authors@R: c(
    person("Yihui", "Xie", role = c("aut", "cre"), email = "xie@yihui.name", comment = c(ORCID = "0000-0003-0645-5666")),
    person("Yixuan", "Qiu", role = "aut"),
    person("Christopher", "Gandrud", role = "ctb"),
    person("Qiang", "Li", role = "ctb")
    )
Description: Provides syntax highlighting for R source code. Currently it
    supports LaTeX and HTML output. Source code of other languages is supported
    via Andre Simon's highlight package (<http://www.andre-simon.de>).
Depends: R (>= 3.2.3)
Imports: xfun (>= 0.18)
Suggests: knitr, markdown, testit
License: GPL
URL: https://github.com/yihui/highr
BugReports: https://github.com/yihui/highr/issues
VignetteBuilder: knitr
RoxygenNote: 7.1.1
NeedsCompilation: no
Packaged: 2021-04-16 21:12:10 UTC; yihui
Author: Yihui Xie [aut, cre] (<https://orcid.org/0000-0003-0645-5666>),
  Yixuan Qiu [aut],
  Christopher Gandrud [ctb],
  Qiang Li [ctb]
Maintainer: Yihui Xie <xie@yihui.name>
Repository: CRAN
Date/Publication: 2021-04-16 21:50:05 UTC

Package: highr
Type: Package
Title: Syntax Highlighting for R Source Code
Version: 0.11
Authors@R: c(
    person("Yihui", "Xie", role = c("aut", "cre"), email = "xie@yihui.name", comment = c(ORCID = "0000-0003-0645-5666")),
    person("Yixuan", "Qiu", role = "aut"),
    person("Christopher", "Gandrud", role = "ctb"),
    person("Qiang", "Li", role = "ctb")
    )
Description: Provides syntax highlighting for R source code. Currently it
    supports LaTeX and HTML output. Source code of other languages is supported
    via Andre Simon's highlight package (<https://gitlab.com/saalen/highlight>).
Depends: R (>= 3.3.0)
Imports: xfun (>= 0.18)
Suggests: knitr, markdown, testit
License: GPL
URL: https://github.com/yihui/highr
BugReports: https://github.com/yihui/highr/issues
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.1
NeedsCompilation: no
Packaged: 2024-05-26 19:27:21 UTC; yihui
Author: Yihui Xie [aut, cre] (<https://orcid.org/0000-0003-0645-5666>),
  Yixuan Qiu [aut],
  Christopher Gandrud [ctb],
  Qiang Li [ctb]
Maintainer: Yihui Xie <xie@yihui.name>
Repository: CRAN
Date/Publication: 2024-05-26 20:00:03 UTC
Built: R 4.4.0; ; 2026-05-12 15:48:43 UTC; unix

# Generated by roxygen2: do not edit by hand

export(hi_andre)
export(hi_html)
export(hi_latex)
export(hilight)
import(utils)
importFrom(xfun,raw_string)
importFrom(xfun,split_source)
importFrom(xfun,valid_syntax)

# highr

[![Build Status](https://travis-ci.org/yihui/highr.svg)](https://travis-ci.org/yihui/highr)
[![Downloads from the RStudio CRAN mirror](https://cranlogs.r-pkg.org/badges/highr)](https://cran.r-project.org/package=highr)

This is an infrastructure R package for syntax highlighting. It supports
LaTeX and HTML output. Not surprisingly, it works best with R code. It
attaches markups onto source code, e.g., it turns

```r
a <- 1 # something
```

into LaTeX code

```latex
\hlstd{a} \hlkwb{<-} \hlnum{1} \hlcom{\# something}
```

or HTML code

```html
<span class="hl std">a</span> <span class="hl kwb">&lt;-</span> <span class="hl num">1</span> <span class="hl com"># something</span>
```

via

```r
library(highr)
hi_latex("a <- 1 # something")
hi_html("a <- 1 # something")
# or hilight(code, format = "latex"/"html")
```

This package also has a wrapper function, `hi_andre()`, for Andre Simon's
[Highlight](http://www.andre-simon.de) package.

There are a few package Markdown vignettes in this package:

```r
browseVignettes(package = "highr")
```

To install the development version here, use

```r
install.packages('highr', repos = 'http://rforge.net', type = 'source')
```

This package is licensed under GPL, and is supposed to serve other packages
such as [**knitr**](http://yihui.name/knitr).

library(testit)

assert('escape_latex() sanitizes backslashes and {}', {
  (escape_latex('\\') %==% '\\textbackslash{}')
  (escape_latex('\\{}') %==% '\\textbackslash{}\\{\\}')
  (escape_latex('{\\}') %==% '\\{\\textbackslash{}\\}')
  (escape_latex('~!@#$%^&*()') %==% '~!@#$%^&*()')
})

assert('escape_html() escapes HTML chars', {
  (escape_html('&"<>') %==% '&amp;&quot;&lt;&gt;')
  (escape_html('~!@#$%^&*()') %==% '~!@#$%^&amp;*()')
})

library(testit)

assert('hi_latex() works without prompts', {
  (hi_latex('1+1') %==% '\\hlnum{1}\\hlopt{+}\\hlnum{1}')
  (hi_latex('  1 +    1') %==% '  \\hlnum{1} \\hlopt{+}    \\hlnum{1}')
  (hi_latex(c('  if (TRUE ){', 'foo && bar}')) %==% c(
    '  \\hlkwa{if} \\hldef{(}\\hlnum{TRUE} \\hldef{)\\{}',
    '\\hldef{foo} \\hlopt{&&} \\hldef{bar\\}}'
  ))
})


assert('hi_latex() works with prompts', {
  (hi_latex('1+1', prompt=TRUE) %==% '\\hldef{> }\\hlnum{1}\\hlopt{+}\\hlnum{1}')
  (hi_latex(c('  if (TRUE ){', 'foo && bar}'), prompt = TRUE) %==% paste(
    '\\hldef{> }  \\hlkwa{if} \\hldef{(}\\hlnum{TRUE} \\hldef{)\\{}',
    '\\hldef{+ }\\hldef{foo} \\hlopt{&&} \\hldef{bar\\}}', sep = '\n'
  ))
})

assert('hi_latex() preserves blank lines', {
  (hi_latex(c('1+1','','foo(x=3) # comm')) %==% c(
    '\\hlnum{1}\\hlopt{+}\\hlnum{1}\n',
    '\\hlkwd{foo}\\hldef{(}\\hlkwc{x}\\hldef{=}\\hlnum{3}\\hldef{)} \\hlcom{# comm}'
  ))
})

assert('the fallback method recognizes comments, functions and strings', {
  (hi_latex('1+1 # a comment', fallback = TRUE) %==% '1+1 \\hlcom{# a comment}')
  (hi_latex('paste("STRING", \'string\')', fallback = TRUE) %==%
            '\\hlkwd{paste}(\\hlsng{"STRING"}, \\hlsng{\'string\'})')
})

assert('the fallback mode is used when the code does not parse', {
  (has_warning(res <- hi_latex('1+1+ # comment')))
  (res %==% '1+1+ \\hlcom{# comment}')
})

assert('hilight() works even if code only contains comments', {
  (hi_latex('# only comments') %==% '\\hlcom{# only comments}')
})

assert('the right arrow -> is preserved', {
  (hi_latex('1 ->x # foo') %==% '\\hlnum{1} \\hlkwb{->}\\hldef{x} \\hlcom{# foo}')
})

assert('blank lines before/after code are preserved', {
  (hi_latex(c('', '', '1')) %==% c('\n', '\\hlnum{1}'))
  (hi_latex(c('', '', '1', '')) %==% c('\n', '\\hlnum{1}', ''))
})

# define one's own markup data frame
my_cmd = cmd_html
my_cmd['NUM_CONST', 1] = '<span class="my num">'

assert('custom markup also works', {
  (hi_html('1+ 1') %==%
    '<span class="hl num">1</span><span class="hl opt">+</span> <span class="hl num">1</span>')
  (hi_html('1+ 1', markup = my_cmd) %==%
    '<span class="my num">1</span><span class="hl opt">+</span> <span class="my num">1</span>')
})

rm(my_cmd)

library(testit)

assert(
  'hi_latex() works without prompts',
  hi_latex('1+1') == '\\hlnum{1}\\hlopt{+}\\hlnum{1}',
  hi_latex('  1 +    1') == '  \\hlnum{1} \\hlopt{+}    \\hlnum{1}',
  identical(hi_latex(c('  if (TRUE ){', 'foo && bar}')), c(
    '  \\hlkwa{if} \\hlstd{(}\\hlnum{TRUE} \\hlstd{)\\{}',
    '\\hlstd{foo} \\hlopt{&&} \\hlstd{bar\\}}'
  ))
)

assert(
  'hi_latex() works with prompts',
  hi_latex('1+1', prompt=TRUE) == '\\hlstd{> }\\hlnum{1}\\hlopt{+}\\hlnum{1}',
  identical(hi_latex(c('  if (TRUE ){', 'foo && bar}'), prompt = TRUE), paste(
    '\\hlstd{> }  \\hlkwa{if} \\hlstd{(}\\hlnum{TRUE} \\hlstd{)\\{}',
    '\\hlstd{+ }\\hlstd{foo} \\hlopt{&&} \\hlstd{bar\\}}', sep = '\n'
  ))
)

assert(
  'hi_latex() preserves blank lines',
  identical(hi_latex(c('1+1','','foo(x=3) # comm')), c(
    '\\hlnum{1}\\hlopt{+}\\hlnum{1}\n',
    '\\hlkwd{foo}\\hlstd{(}\\hlkwc{x}\\hlstd{=}\\hlnum{3}\\hlstd{)} \\hlcom{# comm}'
  ))
)

assert(
  'the fallback method recognizes comments, functions and strings',
  identical(hi_latex('1+1 # a comment', fallback = TRUE), '1+1 \\hlcom{# a comment}'),
  identical(hi_latex('paste("STRING", \'string\')', fallback = TRUE),
            '\\hlkwd{paste}(\\hlstr{"STRING"}, \\hlstr{\'string\'})')
)

assert(
  'the fallback mode is used when the code does not parse',
  has_warning(res <- hi_latex('1+1+ # comment')),
  identical(res, '1+1+ \\hlcom{# comment}')
)

assert(
  'hilight() works even if code only contains comments',
  identical(hi_latex('# only comments'), '\\hlcom{# only comments}')
)

assert(
  'the right arrow -> is preserved',
  identical(hi_latex('1 ->x # foo'), '\\hlnum{1} \\hlkwb{->}\\hlstd{x} \\hlcom{# foo}')
)

assert(
  'blank lines before/after code are preserved',
  hi_latex(c('', '', '1')) %==% c('\n', '\\hlnum{1}'),
  hi_latex(c('', '', '1', '')) %==% c('\n', '\\hlnum{1}', '')
)

# define one's own markup data frame
my_cmd = cmd_html
my_cmd['NUM_CONST', 1] = '<span class="my num">'

assert(
  'custom markup also works',
  hi_html('1+ 1') ==
    '<span class="hl num">1</span><span class="hl opt">+</span> <span class="hl num">1</span>',
  hi_html('1+ 1', markup = my_cmd) ==
    '<span class="my num">1</span><span class="hl opt">+</span> <span class="my num">1</span>'
)

rm(my_cmd)

# parse source and keep source
parse_source = function(lines) parse(text = lines, keep.source = TRUE)

# borrowed from knitr

# escape backslashes and {} for the alltt package
escape_latex = function(x) {
  x = gsub('\\\\', '\\\\textbackslash', x)
  x = gsub('([{}])', '\\\\\\1', x)
  gsub('\\\\textbackslash', '\\\\textbackslash{}', x)
}

# escape special HTML chars
escape_html = function(x) {
  x = gsub('&', '&amp;', x)
  x = gsub('<', '&lt;', x)
  x = gsub('>', '&gt;', x)
  x = gsub('"', '&quot;', x)
  x
}

library(testit)

assert(
  'spaces(n) gives n spaces',
  spaces(-1) == '', spaces(0) == '', spaces(1) == ' ', spaces(5) == '     '
)

assert(
  'escape_latex() sanitizes backslashes and {}',
  escape_latex('\\') == '\\textbackslash{}',
  escape_latex('\\{}') == '\\textbackslash{}\\{\\}',
  escape_latex('{\\}') == '\\{\\textbackslash{}\\}',
  escape_latex('~!@#$%^&*()') == '~!@#$%^&*()'
)

assert(
  'escape_html() escapes HTML chars',
  escape_html('&"<>') == '&amp;&quot;&lt;&gt;',
  escape_html('~!@#$%^&*()') == '~!@#$%^&amp;*()'
)

hi.keywords =  paste('(\\W)(', paste(c(
  'if', 'else', 'repeat', 'while', 'function', 'for', 'in', 'next', 'break', 'repeat',
  'LETTERS', 'letters', 'month.abb', 'month.name', 'pi',
  'TRUE', 'FALSE', 'NULL', 'Inf', 'NaN', 'NA', 'NA_integer_', 'NA_real_', 'NA_complex_', 'NA_character_'
), collapse = '|'), ')(\\W)', sep = '')

#  only highlight function names, strings and comments
hi_naive_latex = function(x, str, com, kwa, kwd) {
  i = grepl('^\\s*#', x)  # whole lines of comments
  x[i] = sprintf(paste(com[1], '%s', com[2], sep = ''), x[i])
  str = escape_bs(str); com = escape_bs(com); kwd = escape_bs(kwd)
  # comments: what if # inside quotes?
  if (any(idx <- grepl('#', x) & !grepl('"', x) & !i))
    x[idx] = gsub('(#.*)', sprintf('%s\\1%s', com[1], com[2]), x[idx])
  i = which(!i)  # not comments
  # function names
  x[i] = gsub('([[:alnum:]_\\.]+)(\\s*)\\(', sprintf('%s\\1%s\\2(', kwd[1], kwd[2]), x[i])
  # character strings
  x[i] = gsub('"([^"]*)"', sprintf('%s"\\1"%s', str[1], str[2]), x[i])
  x[i] = gsub("'([^']*)'", sprintf("%s'\\1'%s", str[1], str[2]), x[i])
  x
}
hi_naive_html = function(x, str, com, kwa, kwd) {
  # character strings
  x = gsub('&quot;(.*?)&quot;', sprintf('%s&quot;\\1&quot;%s', str[1], str[2]), x)
  x = gsub("'([^']*)'", sprintf("%s'\\1'%s", str[1], str[2]), x)
  # function names
  x = gsub('([[:alnum:]_\\.]+)(\\s*)\\(', sprintf('%s\\1%s\\2(', kwd[1], kwd[2]), x)
  if (any(idx <- grepl('#', x) & !grepl('"', x)))
    x[idx] = gsub('(#.*)', sprintf('%s\\1%s', com[1], com[2]), x[idx])
  gsub(hi.keywords, sprintf('\\1%s\\2%s\\3', kwa[1], kwa[2]), x)
}

hi_naive = function(code, format = c('latex', 'html'), markup, escape_fun = identity,
                    prompt = NULL) {
  format = match.arg(format)
  code = escape_fun(code)
  if (length(prompt) == 2) {
    # borrowed from knitr:::line_prompt
    code = paste(
      prompt[1],
      gsub('(?<=\n)(?=.|\n)', prompt[2], code, perl = TRUE), sep = ''
    )
  }
  hi_fun = if (format == 'latex') hi_naive_latex else hi_naive_html
  hi_fun(split_lines(code), str = markup['STR_CONST', ], com = markup['COMMENT', ],
         kwa = markup['IF', ], kwd = markup['SYMBOL_FUNCTION_CALL', ])
}

# split a character vector by \n
split_lines = function(x) {
  if (!any(grepl('\n', x))) return(x)
  x[x == ''] = '\n'
  unlist(strsplit(x, '\n'))
}

# escape backslashes
escape_bs = function(x) gsub('\\\\', '\\\\\\\\', x)

# some special symbols (keywords a)
.keywords = c('FUNCTION', 'IF', 'ELSE', 'WHILE', 'FOR', 'IN', 'BREAK', 'REPEAT', 'NEXT', 'NULL_CONST')

.operators = c(
  sprintf("'%s'", c('+', '-', '*', '/', '^', '$', '@', ':', '?', '~', '!')),
  'SPECIAL', 'GT', 'GE', 'LT', 'LE', 'EQ', 'NE', 'AND', 'AND2', 'OR', 'OR2',
  'NS_GET', 'NS_GET_INT'
)

.cmd.list = sort(c(
  NUM_CONST            = 'num', # numbers
  SYMBOL_FUNCTION_CALL = 'kwd', # function calls
  STR_CONST            = 'sng', # character strings
  COMMENT              = 'com', # comment
  SYMBOL_FORMALS       = 'kwc', # function(formals)
  SYMBOL_SUB           = 'kwc', # FUN(args)
  SLOT                 = 'kwc', # S4 slot
  LEFT_ASSIGN          = 'kwb', # assignment
  EQ_ASSIGN            = 'kwb',
  RIGHT_ASSIGN         = 'kwb',
  setNames(rep('opt', length(.operators)), .operators),
  setNames(rep('kwa', length(.keywords)), .keywords),
  DEFAULT              = 'def' # everything else
))

cmd_latex = data.frame(
  cmd1 = paste('\\hl', .cmd.list, '{', sep = ''),
  cmd2 = '}',
  stringsAsFactors = FALSE,
  row.names = names(.cmd.list)
)

cmd_latex = data.frame(
  cmd1 = paste('\\hl', .cmd.list, '{', sep = ''),
  cmd2 = '}',
  stringsAsFactors = FALSE,
  row.names = names(.cmd.list)
)

.cmd.pandoc = c(
  num = 'DecValTok', kwd = 'KeywordTok', sng = 'StringTok', com = 'CommentTok',
  kwc = 'DataTypeTok', kwb = 'NormalTok', opt = 'NormalTok', kwa = 'NormalTok',
  def = 'NormalTok'
)

cmd_pandoc_latex = data.frame(
  cmd1 = paste('\\', .cmd.pandoc[.cmd.list], '{', sep = ''),
  cmd2 = '}',
  stringsAsFactors = FALSE,
  row.names = names(.cmd.list)
)

cmd_html = data.frame(
  cmd1 = paste('<span class="hl ', .cmd.list, '">', sep = ''),
  cmd2 = '</span>',
  stringsAsFactors = FALSE,
  row.names = names(.cmd.list)
)

# merge code and markups; use standard markups on unknown tokens
merge_cmd = function(pdata, cmd) {
  res = cmd[pdata$token, ]
  idx = is.na(res[, 1])
  res[idx, 1] = cmd['DEFAULT', 1]
  res[idx, 2] = cmd['DEFAULT', 2]
  res[is.na(res)] = '' # if DEFAULT is undefined in the markup data frame
  res
}

#' Syntax highlight an R code fragment
#'
#' This function \code{\link{parse}}s the R code, fetches the tokens in it
#' (\code{\link{getParseData}}), and attach syntax highlighting commands onto
#' them. With proper style definitions for these commands (such as colors or
#' font styles), the R code will be syntax highlighted in the LaTeX/HTML output.
#' The two functions \code{hi_latex} and \code{hi_html} are wrappers of
#' \code{hilight} for LaTeX and HTML output, respectively.
#'
#' For the \code{markup} data frame, the first column is put before the R
#' tokens, and the second column is behind; the row names of the data frame must
#' be the R token names; a special row is named \code{DEFAULT}, which contains
#' the markup for the standard tokens (i.e. those that do not need to be
#' highlighted); if missing, the built-in data frames \code{highr:::cmd_latex}
#' and \code{highr:::cmd_html} will be used.
#'
#' This function only binds markups onto R tokens, and the real syntax
#' highlighting must be done with style definitions, which is out of the scope
#' of this package. It was designed to be used as the syntax highlighting
#' infrastructure of other packages such as \pkg{knitr}, where the colors and
#' font styles are properly defined in the LaTeX preamble and HTML header.
#' @param code a character string (the R source code)
#' @param format the output format
#' @param markup a data frame of two columns containing the markup commands
#' @param prompt whether to add prompts to the code
#' @param fallback whether to use the fallback method, i.e. the regular
#'   expression based method; this method is not precise and only highlights a
#'   few types of symbols such as comments, strings and functions;
#'   \code{fallback} will be set to \code{TRUE} when the input \code{code} fails
#'   to be \code{\link{parse}d}
#' @param ... arguments to be passed to \code{hilight()}
#' @author Yihui Xie and Yixuan Qiu
#' @seealso See the package vignettes \code{browseVignettes('highr')} for how
#' this function works internally.
#' @return A character vector for the syntax highlighted code.
#' @examples library(highr)
#' hilight("x=1 # assignment")
#'
#' txt = c("a <- 1 # something", 'c(y="world", z="hello")', 'b=function(x=5) {',
#' 'for(i in 1:10) {
#'   if (i < x) print(i) else break}}',
#' "z@@child # S4 slot", "'special chars <>#$%&_{}'")
#' cat(hi_latex(txt), sep = '\n')
#' cat(hi_html(txt), sep = '\n')
#'
#' # the markup data frames
#' highr:::cmd_latex; highr:::cmd_html
#' @import utils
#' @export
hilight = function(code, format = c('latex', 'html'), markup, prompt = FALSE, fallback = FALSE) {
  if (length(code) == 0) return(code)
  format = match.arg(format)
  if (missing(markup) || is.null(markup))
    markup = if (format == 'latex') cmd_latex else cmd_html
  escape_fun = if (format == 'latex') escape_latex else escape_html
  if (!fallback && !xfun::valid_syntax(code, silent = FALSE)) {
    # the code is not valid, so you must use the fallback mode
    warning('the syntax of the source code is invalid; the fallback mode is used')
    fallback = TRUE
  }
  if (!prompt) return(
    (if (fallback) hi_naive else hilight_one)(code, format, markup, escape_fun)
  )
  p1 = escape_fun(getOption('prompt')); p2 = escape_fun(getOption('continue'))
  std = unlist(markup['DEFAULT', ])
  if (!any(is.na(std))) {
    p1 = paste0(std[1], p1, std[2]); p2 = paste0(std[1], p2, std[2])
  }
  code = xfun::split_source(code)
  sapply(mapply(hilight_one, code, MoreArgs = list(format, markup, escape_fun),
                SIMPLIFY = FALSE, USE.NAMES = FALSE),
         function(x) paste0(rep(c(p1, p2), c(1L, length(x) - 1L)), x, collapse = '\n'))
}
# highlight one expression
hilight_one = function(code, format, markup, escape_fun) {

  # the data frames do not need factors in this function; need to keep source
  op = options(stringsAsFactors = FALSE, keep.source = TRUE); on.exit(options(op))

  p = parse_source(code)
  z = utils::getParseData(p)
  if (NROW(z) == 0L || !any(z$terminal)) return(code)
  z = z[z$terminal, ]

  # record how empty lines before/after the code
  one = paste(code, collapse = '\n')
  r1 = '^(\\s*)\n.*'; r2 = '^.*?\n(\\s*)$'
  s1 = if (grepl(r1, one)) gsub(r1, '\\1', one)
  s2 = if (grepl(r2, one)) gsub(r2, '\\1', one)

  res = cbind(z[, c('line1', 'col1', 'line2', 'col2', 'text')], merge_cmd(z, markup))

  # escape special LaTeX/HTML chars
  res$text = escape_fun(res$text)

  # record how many blank lines after each token
  blanks = c(pmax(res$line1[-1] - res$line2[-nrow(res)] - 1, 0), 0)
  # add line breaks to the 8th column
  res = cbind(res, strrep('\n', blanks))

  # e.g. a string spans across multiple lines; now need to replace line1 with
  # line2 so that we know the starting and ending positions of spaces; e.g. turn
  # line/col numbers 1 5 2 6 into 2 5 2 6
  for (i in which(res$line1 != res$line2)) {
    res$line1[res$line1 == res$line1[i]] = res$line2[i]
  }

  out = lapply(split(res, res$line1), function(d) {
    # merge adjacent tokens of the same type so that the output is cleaner
    empty = matrix(FALSE, nrow = nrow(d), ncol = 2)
    for (i in seq_len(nrow(d) - 1)) {
      if (all(d[i, 6:7] == d[i + 1, 6:7])) empty[i + 1, 1] = empty[i, 2] = TRUE
    }
    d[, 6:7][empty] = ''
    col = as.matrix(d[, c('col1', 'col2')])
    # add 0 and remove col[n, 2] to get start/end positions of spaces
    col = matrix(head(c(0, t(col)), -1), ncol = 2, byrow = TRUE)
    paste(strrep(' ', col[, 2] - col[, 1] - 1), d[, 6], d[, 'text'], d[, 7],
          d[, 8], sep = '', collapse = '')
  })
  c(s1, unlist(out, use.names = FALSE), s2)
}
#' @export
#' @rdname hilight
hi_latex = function(code, ...) hilight(code, 'latex', ...)
#' @export
#' @rdname hilight
hi_html = function(code, ...) hilight(code, 'html', ...)

#' A wrapper to Andre Simon's Highlight
#'
#' This function calls Highlight to syntax highlight a code fragment.
#' @param code a character string of the source code
#' @param language the input language (c, cpp, python, r, ...); see
#'   \code{system('highlight -p')}
#' @param format the output format (html, latex, ...)
#' @references Andre Simon's Highlight package
#'   \url{https://gitlab.com/saalen/highlight}.
#' @return A character string for the syntax highlighted code.
#' @export
#' @examples \dontrun{hi_andre('1+1', language='R')
#' hi_andre('void main() {\nreturn(0)\n}', language='c', format='latex')}
hi_andre = function(code, language, format = 'html') {
  h = Sys.which('highlight')
  os = Sys.info()[['sysname']]
  # highlight on Linux Mint can be something else
  # on OS10 with highlight installed using Homebrew it's often in /usr/local/bin
  if (!nzchar(h) || (h == '/usr/local/bin/highlight' && os != 'Darwin' &&
                       !file.exists(h <- '/usr/bin/highlight')))
    stop('please first install highlight from https://gitlab.com/saalen/highlight')
  f = basename(tempfile('code', '.'))
  writeLines(code, f); on.exit(unlink(f))
  cmd = sprintf('%s -f -S %s -O %s %s', shQuote(h), correct_lang(language), format, f)
  system(cmd, intern = TRUE)
}

# to help knitr engines decide the highlight language
correct_lang = function(x) {
  switch(x, Rcpp = 'cpp', tikz = 'latex', Rscript = 'R', fortran = 'f', stan = 'R', x)
}

<!--
%\VignetteEngine{knitr::docco_classic}
%\VignetteIndexEntry{Internals of the highr package}
-->

# Internals of the `highr` package

The **highr** package is based on the function `getParseData()`, which was
introduced in R 3.0.0. This function gives detailed information of the
symbols in a code fragment. A simple example:

```{r}
p = parse(text = "   xx = 1 + 1  # a comment", keep.source = TRUE)
(d = getParseData(p))
```

The first step is to filter out the rows that we do not need:

```{r}
(d = d[d$terminal, ])
```

There is a column `token` in the data frame, and we will wrap this column
with markup commands, e.g. `\hlnum{1}` for the numeric constant `1`. We
defined the markup commands in `cmd_latex` and `cmd_html`:

```{r}
head(highr:::cmd_latex)
tail(highr:::cmd_html)
```

These command data frames are connected to the tokens in the R code via
their row names:

```{r}
d$token
rownames(highr:::cmd_latex)
```

Now we know how to wrap up the R tokens. The next big question is how to
restore the white spaces in the source code, since they were not directly
available in the parsed data, but the parsed data contains column numbers,
and we can derive the positions of white spaces from them. For example,
`col2 = 5` for the first row, and `col1 = 7` for the next row, and that
indicates there must be one space after the token in the first row, otherwise
the next row will start at the position `6` instead of `7`.

A small trick is used to fill in the gaps of white spaces:

```{r}
(z = d[, c('col1', 'col2')])  # take out the column positions
(z = t(z)) # transpose the matrix
(z = c(z)) # turn it into a vector
(z = c(0, head(z, -1))) # append 0 in the beginning, and remove the last element
(z = matrix(z, ncol = 2, byrow = TRUE))
```

Now the two columns indicate the starting and ending positions of spaces,
and we can easily figure out how many white spaces are needed for each row:

```{r}
(s = z[, 2] - z[, 1] - 1)
(s = strrep(' ', s))
paste(s, d$text, sep = '')
```

So we have successfully restored the white spaces in the source code. Let's
paste all pieces together (suppose we highlight for LaTeX):

```{r}
m = highr:::cmd_latex[d$token, ]
cbind(d, m)
# use standard markup if tokens do not exist in the table
m[is.na(m[, 1]), ] = highr:::cmd_latex['DEFAULT', ]
paste(s, m[, 1], d$text, m[, 2], sep = '', collapse = '')
```

So far so simple. That is one line of code, after all. A next challenge
comes when there are multiple lines, and a token spans across multiple lines:

```{r}
d = getParseData(parse(text = "x = \"a character\nstring\" #hi", keep.source = TRUE))
(d = d[d$terminal, ])
```

Take a look at the third row. It says that the character string starts from
line 1, and ends on line 2. In this case, we just pretend as if everything
on line 1 were on line 2. Then for each line, we append the missing spaces
and apply markup commands to text symbols.

```{r}
d$line1[d$line1 == 1] = 2
d
```

Do not worry about the column `line2`. It does not matter. Only `line1` is
needed to indicate the line number here.

Why do we need to highlight line by line instead of applying highlighting
commands to all text symbols (a.k.a vectorization)? Well, the margin of this
paper is too small to write down the answer.

<!--
%\VignetteEngine{knitr::docco_classic}
%\VignetteIndexEntry{Customization of the highr package}
-->

# Customization of the `highr` package

If you are not satisfied with the default syntax highlighting commands in
the **highr** package, you can just use your own tags/commands. In this
vignette, we show a brief example.

The default highlighting commands are stored in two internal data frames
`cmd_latex` and `cmd_html`:

```{r}
library(highr)
highr:::cmd_latex
```

This data frame is passed to the `markup` argument in `hilight()`, so we are
free to pass a modified version there. Suppose I want to use the command
`\my<*>` instead of `\hl<*>`:

```{r}
m = highr:::cmd_latex
m[, 1] = sub('\\hl', '\\my', m[, 1], fixed = TRUE)
head(m)
```

Then

```{r}
hilight("x = 1+1  # a comment")  # default markup
hilight("x = 1+1  # a comment", markup = m)  # custom markup
```

This allows one to use arbitrary commands around the text symbols in the R
code. See <https://github.com/yihui/highr/blob/master/R/highlight.R> for how
`cmd_latex` and `cmd_html` were generated in **highr**.

# hi_latex() works without prompts

```{r}
hi_latex('1+1')
hi_latex('  1 +    1')
hi_latex(c('  if (TRUE ){', 'foo && bar}'))
```
```
\hlnum{1}\hlopt{+}\hlnum{1}
  \hlnum{1} \hlopt{+}    \hlnum{1}
  \hlkwa{if} \hldef{(}\hlnum{TRUE} \hldef{)\{}
\hldef{foo} \hlopt{&&} \hldef{bar\}}
```

# hi_latex() works with prompts

```{r}
hi_latex('1+1', prompt = TRUE)
hi_latex(c('  if (TRUE ){', 'foo && bar}'), prompt = TRUE)
```
```
\hldef{> }\hlnum{1}\hlopt{+}\hlnum{1}
\hldef{> }  \hlkwa{if} \hldef{(}\hlnum{TRUE} \hldef{)\{}
\hldef{+ }\hldef{foo} \hlopt{&&} \hldef{bar\}}
```

# hi_latex() preserves blank lines

```{r}
hi_latex(c('1+1', '', 'foo(x=3) # comm'))
```
```
\hlnum{1}\hlopt{+}\hlnum{1}

\hlkwd{foo}\hldef{(}\hlkwc{x}\hldef{=}\hlnum{3}\hldef{)} \hlcom{# comm}
```

# the fallback method recognizes comments, functions and strings

```{r}
hi_latex('1+1 # a comment', fallback = TRUE)
hi_latex('paste("STRING", \'string\')', fallback = TRUE)
```
```
1+1 \hlcom{# a comment}
\hlkwd{paste}(\hlsng{"STRING"}, \hlsng{'string'})
```

# the fallback mode is used when the code does not parse

```{r}
has_warning(res <- hi_latex('1+1+ # comment'))
res
```
```
[1] TRUE
1+1+ \hlcom{# comment}
```

# hilight() works even if code only contains comments

```{r}
hi_latex('# only comments')
```
```
\hlcom{# only comments}
```

# the right arrow -> is preserved

```{r}
hi_latex('1 ->x # foo')
```
```
\hlnum{1} \hlkwb{->}\hldef{x} \hlcom{# foo}
```

# blank lines before/after code are preserved

```{r}
hi_latex(c('', '', '1'))
hi_latex(c('', '', '1', ''))
```
```


\hlnum{1}


\hlnum{1}

```

# custom markup also works

```{r}
hi_html('1+ 1')
my_cmd = cmd_html
my_cmd['NUM_CONST', 1] = '<span class="my num">'
hi_html('1+ 1', markup = my_cmd)
rm(my_cmd)
```
```
<span class="hl num">1</span><span class="hl opt">+</span> <span class="hl num">1</span>
<span class="my num">1</span><span class="hl opt">+</span> <span class="my num">1</span>
```

# escape_latex() sanitizes backslashes and {}

```{r}
escape_latex('\\')
escape_latex('\\{}')
escape_latex('{\\}')
escape_latex('~!@#$%^&*()')
```
```
\textbackslash{}
\textbackslash{}\{\}
\{\textbackslash{}\}
~!@#$%^&*()
```

# escape_html() escapes HTML chars

```{r}
escape_html('&"<>')
escape_html('~!@#$%^&*()')
```
```
&amp;&quot;&lt;&gt;
~!@#$%^&amp;*()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/highlight.R
\name{hilight}
\alias{hilight}
\alias{hi_latex}
\alias{hi_html}
\title{Syntax highlight an R code fragment}
\usage{
hilight(
  code,
  format = c("latex", "html"),
  markup,
  prompt = FALSE,
  fallback = FALSE
)

hi_latex(code, ...)

hi_html(code, ...)
}
\arguments{
\item{code}{a character string (the R source code)}

\item{format}{the output format}

\item{markup}{a data frame of two columns containing the markup commands}

\item{prompt}{whether to add prompts to the code}

\item{fallback}{whether to use the fallback method, i.e. the regular
expression based method; this method is not precise and only highlights a
few types of symbols such as comments, strings and functions;
\code{fallback} will be set to \code{TRUE} when the input \code{code} fails
to be \code{\link{parse}d}}

\item{...}{arguments to be passed to \code{hilight()}}
}
\value{
A character vector for the syntax highlighted code.
}
\description{
This function \code{\link{parse}}s the R code, fetches the tokens in it
(\code{\link{getParseData}}), and attach syntax highlighting commands onto
them. With proper style definitions for these commands (such as colors or
font styles), the R code will be syntax highlighted in the LaTeX/HTML output.
The two functions \code{hi_latex} and \code{hi_html} are wrappers of
\code{hilight} for LaTeX and HTML output, respectively.
}
\details{
For the \code{markup} data frame, the first column is put before the R
tokens, and the second column is behind; the row names of the data frame must
be the R token names; a special row is named \code{STANDARD}, which contains
the markup for the standard tokens (i.e. those that do not need to be
highlighted); if missing, the built-in data frames \code{highr:::cmd_latex}
and \code{highr:::cmd_html} will be used.

This function only binds markups onto R tokens, and the real syntax
highlighting must be done with style definitions, which is out of the scope
of this package. It was designed to be used as the syntax highlighting
infrastructure of other packages such as \pkg{knitr}, where the colors and
font styles are properly defined in the LaTeX preamble and HTML header.
}
\examples{
library(highr)
hilight("x=1 # assignment")

txt = c("a <- 1 # something", "c(y=\"world\", z=\"hello\")", "b=function(x=5) {", 
    "for(i in 1:10) {
  if (i < x) print(i) else break}}", "z@child # S4 slot", 
    "'special chars <>#$\%&_{}'")
cat(hi_latex(txt), sep = "\n")
cat(hi_html(txt), sep = "\n")

# the markup data frames
highr:::cmd_latex
highr:::cmd_html
}
\seealso{
See the package vignettes \code{browseVignettes('highr')} for how
this function works internally.
}
\author{
Yihui Xie and Yixuan Qiu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/highlight.R
\name{hi_andre}
\alias{hi_andre}
\title{A wrapper to Andre Simon's Highlight}
\usage{
hi_andre(code, language, format = "html")
}
\arguments{
\item{code}{a character string of the source code}

\item{language}{the input language (c, cpp, python, r, ...); see
\code{system('highlight -p')}}

\item{format}{the output format (html, latex, ...)}
}
\value{
A character string for the syntax highlighted code.
}
\description{
This function calls Highlight to syntax highlight a code fragment.
}
\examples{
\dontrun{
hi_andre("1+1", language = "R")
hi_andre("void main() {\nreturn(0)\n}", language = "c", format = "latex")
}
}
\references{
Andre Simon's Highlight package
  \url{https://gitlab.com/saalen/highlight}.
}
